import panel as pn

# metastore database configurations
pg_username = 'postgres'
pg_password = 'PiIs&&&31415926535'
pg_server = 'localhost'
pg_database = 'postgres'
pg_port = '5432'

# spark configurations
spark_warehouse = '/home/morteza/Documents/Projects/letme/storage/spark-warehouse'

# max resources for each app
spark_max_memory = 20
spark_max_cpu_cores = 8

# config panel
def pn_setup():
    pn.extension('tabulator')
    pn.widgets.Tabulator.theme = 'simple'





from pyspark.sql import SparkSession
from delta import *
import pandas as pd
import getpass
import panel as pn

from .config import *
from .auth import *
from .helpers import *
from .gui import *
from .source.mssql import MSSQL

pn_setup()

class App:
    def __init__(self, username: str, password: str = None, spark_memory: int = 2, spark_cpu_cores: int = 1):
        self.spark_memory = spark_max_memory if spark_memory > spark_max_memory else spark_memory
        self.spark_cpu_cores = spark_max_cpu_cores if spark_cpu_cores > spark_max_cpu_cores else spark_cpu_cores
        self.username = username
        self.__password = getpass.getpass('Please enter your password: ') if not password else password
        self.user_id = None
        self.__spark = None
        self.__start()

        self.read = self.__spark.read
        self.createDataFrame = self.__spark.createDataFrame
        self.stop = self.__spark.stop

    def __start(self):
        if not Auth(self.username, self.__password).login():
            print('The username and/or password is incorrect. Make sure you have an account and it is activated.')
            return None
        self.spark_memory = f'{self.spark_memory}g'
        self.spark_cpu_cores = f'{self.spark_cpu_cores}'
        self.__set_user_id()
        self.__spark = self.__create_spark_session()

    def __set_user_id(self):
        sql = f''' select user_id from users where user_username = '{self.username}' and user_password = '{self.__password}'  '''
        self.user_id = metastore_select(sql)[0][0]

    def __create_spark_session(self):
        builder = SparkSession.builder.master(f'local[{self.spark_cpu_cores}]') \
            .config('spark.sql.extensions', 'io.delta.sql.DeltaSparkSessionExtension') \
            .config('spark.sql.catalog.spark_catalog', 'org.apache.spark.sql.delta.catalog.DeltaCatalog') \
            .config('spark.sql.warehouse.dir', spark_warehouse) \
            .config('spark.driver.memory', self.spark_memory) \
            .config('spark.driver.maxResultSize', self.spark_memory) \
            .config('spark.sql.repl.eagerEval.enabled', True) \
            .config('spark.databricks.delta.schema.autoMerge.enabled', True) \
            .config('spark.databricks.delta.autoCompact.enabled', True)
        return builder.getOrCreate()

    #######################################################################################################
    ## General methods
    #######################################################################################################

    def sql(self, script: str, gui: bool = False):
        try:
            output = self.__spark.sql(script)
        except:
            return 'Not found.'
        if identify_script_type(script) != 'other_script':
            metastore_save_script(self, script)

        df = output.toPandas()
        if not df.empty:
            if gui:
                return simple(output.toPandas())
            return output

    def load(self):
        rows = Auth(self.username, self.__password).get_metastore_scripts()
        for row in rows:
            try:
                self.__spark.sql(row[0])
            except:
                pass

    def list_objects(self, gui: bool = False):
        sql = f'''
        select distinct
            object_name,
            object_type,
            object_created_at
        from objects o inner join objects_users ou on o.object_id = ou.object_id 
        where ou.user_id = {self.user_id}
        '''
        df = pd.read_sql(sql, get_metastore_con())
        if gui:
            return simple(df)
        return df

    def drop_object(self, object_name: str):
        sql = f'''
        select object_id, object_type from objects o where object_name = '{object_name}' and object_owner_id = {self.user_id}
        '''
        row = metastore_select(sql)
        if not row:
            print('Object not found.')
            return
        object_id = row[0][0]
        object_type = row[0][1]

        try:
            self.__spark.sql(f'''drop {object_type} {object_name}''')
        except:
            pass
        sql = f'''
        delete from objects where object_id = {object_id}
        '''
        metastore_ciud(sql)

        sql = f'''
        delete from objects_users where object_id = {object_id}
        '''
        metastore_ciud(sql)

    #######################################################################################################
    ## Permissions and Access
    #######################################################################################################

    def give_permission(self, target_username: str, object_name: str):
        sql = f'''
        select
            distinct o.object_id
        from objects o inner join objects_users ou on o.object_id = ou.object_id
        where 
            o.object_owner_id = {self.user_id} and o.object_name = '{object_name}'
        '''
        row = metastore_select(sql)
        if row:
            object_id = row[0][0]
        else:
            print('You are not the owner of the Object and can not give permission on this object.')
            return

        sql = f''' select user_id from users u where user_username = '{target_username}' '''
        row = metastore_select(sql)
        if row:
            target_user_id = row[0][0]
        else:
            print('Target username not found!')
            return

        sql = f''' insert into objects_users (object_id, user_id) values ({object_id}, {target_user_id}) '''
        try:
            metastore_ciud(sql)
        except:
            pass

    def remove_permission(self, target_username: str, object_name: str):
        sql = f'''
        select * from objects where object_owner_id = {self.user_id} and object_name = '{object_name}' 
        '''
        row = metastore_select(sql)
        if not row:
            print('You are not the owner of this object and can not remove permission.')
            return

        sql = f'''
        select
            distinct ou.object_user_id
        from objects o inner join objects_users ou on o.object_id = ou.object_id
        inner join users u on u.user_id = ou.user_id 
        where 
            u.user_username = '{target_username}' and o.object_name = '{object_name}'
        '''
        row = metastore_select(sql)
        if row:
            sql = f'''delete from objects_users where object_user_id = {row[0][0]}'''
            metastore_ciud(sql)

    #######################################################################################################
    ## Add sources
    #######################################################################################################

    def add_mssql_source(
            self,
            source_name: str,
            server: str,
            database: str,
            username: str,
            password: str,
            port: int = 1433
    ):
        return MSSQL(
            app=self,
            source_name=source_name,
            server=server,
            database=database,
            username=username,
            password=password,
            port=port
        )

    #######################################################################################################
    ## Work with sources
    #######################################################################################################

    def find_source(self, source_name_like:str, gui: bool = False):
        sql = f'''
        select distinct
            object_name as source_name,
            object_created_at 
        from objects o inner join objects_users ou on ou.object_id = o.object_id 
        where o.object_type = 'source' and ou.user_id = {self.user_id} and o.object_name like '%{source_name_like}%' 
        '''
        df = pd.DataFrame(metastore_select(sql))
        if df.empty:
            return
        if gui:
            return simple(df)
        return df

    def load_source(self, source_name: str):
        sql = f'''
        select
            object_extra_info 
        from objects o inner join objects_users ou on ou.object_id = o.object_id 
        where o.object_type = 'source' and ou.user_id = 1 and o.object_name = '{source_name}'
        '''
        output = metastore_select(sql)
        if not output:
            print('Source not found.')
            return
        object_extra_info = output[0][0]
        if object_extra_info.get('source_type') == 'mssql':
            return self.add_mssql_source(
                source_name=object_extra_info.get('source_name'),
                server=object_extra_info.get('server'),
                database=object_extra_info.get('database'),
                username=object_extra_info.get('username'),
                password=object_extra_info.get('password'),
                port=object_extra_info.get('port')
            )



from sqlalchemy import create_engine, text
from pathlib import Path
import re
from datetime import datetime

from .config import *

# get metastore sqlalchemy engine
def get_metastore_con():
    connection_string = f'postgresql://{pg_username}:{pg_password}@{pg_server}:{pg_port}/{pg_database}'
    return create_engine(connection_string, isolation_level='autocommit')

def metastore_ciud(script: str):
    engine = get_metastore_con()
    with engine.connect() as conn:
        conn.execute(text(script))

def metastore_select(script: str):
    engine = get_metastore_con()
    with engine.connect() as conn:
        return conn.execute(text(script)).fetchall()

# normalize the script
def normalize_script(script: str):
    script = script.replace('\n', ' ').strip().replace(';', '').replace('"', "'")
    new_words = [word.strip() for word in script.split(' ') if len(word.strip()) != 0]
    script = ' '.join(new_words)
    return f'{script} ;'

# get absolute path related to a given path
def get_absolute_path(system_path: str):
    return Path(system_path).resolve().__str__()

# create metastore entities
def initdb():
    sql = '''
    create table if not exists objects 
    ( 
        object_id serial primary key,
        object_type varchar(256),
        object_name varchar(256),
        object_owner_id int,
        object_location varchar(256),
        object_script text,
        object_extra_info json,
        object_created_at timestamp
    );
    
    create unique index if not exists objects_unique_type_name on objects (object_type, object_name);
    
    create unique index if not exists objects_unique_location on objects (object_location) where length(object_location) > 0;
    
    create table if not exists users 
    (
        user_id serial primary key, 
        user_username varchar(256), 
        user_password varchar(256),
        user_email varchar(256),
        user_is_active boolean default false
    );
    
    create unique index if not exists users_unique_username on users (user_username);
    
    insert into users (user_username, user_password, user_email, user_is_active) values ('admin', 'admin1234', '', True);
    
    insert into users (user_username, user_password, user_email, user_is_active) values ('testuser', 'testuser1234', '', True);
    
    
    create table if not exists objects_users 
    ( 
        object_user_id serial primary key, 
        user_id int, 
        object_id int 
    );
    
    create unique index if not exists objects_users_unique_persmission on objects_users (object_id, user_id);
    '''
    try:
        metastore_ciud(sql)
    except:
        pass

# identity the type of script user has entered
def identify_script_type(script: str):
    script = normalize_script(script).lower()
    patterns = {
        'create_table': r'^create\s(or replace\s)?table\s(if not exists\s)?.*\susing\sdelta\slocation\s.*\s;$',
        'create_view': r'^create\s(or replace\s)?view\s(if not exists\s)?.*\sas\sselect\s.*\s;$',
        'create_database': r'^create\sdatabase\s(if not exists\s)?.*\s;$'
    }
    for key, value in patterns.items():
        if re.match(value, script):
            return key
    return 'other_script'

# extract info from create table script
def __extract_info_from_create_table_script(script: str):
    start = script.lower().find(' using delta location ') + 21
    end = len(script.lower())
    table_location = script[start:end].replace(';', '').strip().replace("'", "")
    table_location = get_absolute_path(table_location)
    if ' if not exists ' in script.lower():
        start = script.lower().find(' if not exists ') + 14
    else:
        start = script.lower().find(' table ') + 6
    end = script.find('(')
    table_name = script[start:end].strip()
    return {
        'object_script': f'''{script}''',
        'object_type': 'table',
        'object_name': table_name,
        'object_location': table_location,
        'object_created_at': datetime.now()
    }

# extract info from create view script
def __extract_info_from_create_view_script(script: str):
    if ' if not exists ' in script.lower():
        start = script.lower().find(' if not exists ') + 14
    else:
        start = script.lower().find(' view ') + 5
    end = script.lower().find('as')
    view_name = script[start:end].strip()
    return {
        'object_script': f'''{script}''',
        'object_type': 'view',
        'object_name': view_name,
        'object_location': '',
        'object_created_at': datetime.now()
    }

# extract info from create database script
def __extract_info_from_create_database_script(script: str):
    db_name = script.replace(';', '').strip().split(' ')[-1]
    return {
        'object_script': f'''{script}''',
        'object_type': 'database',
        'object_name': db_name,
        'object_location': '',
        'object_created_at': datetime.now()
    }

# extract the info needed to save in the metastore
def __extract_info_from_user_script(script: str):
    script = normalize_script(script)
    script_type = identify_script_type(script)
    if script_type == 'create_table':
        return __extract_info_from_create_table_script(script)
    elif script_type == 'create_view':
        return __extract_info_from_create_view_script(script)
    elif script_type == 'create_database':
        return __extract_info_from_create_database_script(script)

# save the object to the metastore
def metastore_save_script(app ,script: str):
    info = __extract_info_from_user_script(script)
    data = [{
        'object_script': str(info.get('object_script')),
        'object_type': info.get('object_type'),
        'object_name': info.get('object_name'),
        'object_location': info.get('object_location'),
        'object_owner_id': app.user_id,
        'object_created_at': str(info.get('object_created_at'))
    }]
    sql = '''
         insert into public.objects (object_script, object_type, object_name, object_location, object_owner_id, object_created_at)
         values (:object_script, :object_type, :object_name, :object_location, :object_owner_id, :object_created_at) returning object_id
         '''
    engine = get_metastore_con()
    with engine.connect() as conn:
        object_id = conn.execute(text(sql), data).fetchone()[0]

    sql = f''' insert into objects_users (object_id, user_id) values ({object_id}, {app.user_id})'''
    metastore_ciud(sql)

def register_user():
    pass


from .helpers import metastore_select

class Auth:
    def __init__(self, username: str, password: str):
        self.username = username
        self.password = password

    def login(self):
        sql = f''' select * from users where user_username = '{self.username}' and user_password = '{self.password}' and user_is_active is True '''
        rows = metastore_select(sql)
        if not rows:
            return False
        return True

    def get_metastore_scripts(self):
        sql = f'''
        select 
            o.object_script 
        from objects o inner join objects_users ou on o.object_id = ou.object_id 
        inner join users u on ou.user_id = u.user_id 
        where 
            u.user_username = '{self.username}' and u.user_password = '{self.password}' and object_script is not null
        order by o.object_created_at asc
        '''
        return metastore_select(sql)


import panel as pn

def simple(df):
    tabulator_editors = {}
    for column in df.columns:
        tabulator_editors[f'{column}'] = None

    return pn.widgets.Tabulator(
        df,
        layout='fit_columns',
        header_align='center',
        text_align='left',
        pagination='remote',
        page_size=5,
        sizing_mode='stretch_width',
        show_index=False,
        selectable=False,
        editors=tabulator_editors
    )


from setuptools import setup, find_packages

setup(
    name='letme',
    version='1.0',
    author='hamodirtyeza',
    author_email='hamodirtyeza97@gmail.com',
    description='A package to simplify working with Delta Lake and Spark-SQL',
    packages=find_packages(include=['letme',], exclude=[]),
    python_requires='>=3.12',
)